// This program gets all the node tags and the coordinates of the model.
//
// How to build/run? (windows) 
//   [in example/gmsh_api]
//   mkdir build
//   cd build
//   cmake .. && make && code6_allnodes.exe ..\rectangle.geo

#include <gmsh.h>
#include <iostream>

int main(int argc, char **argv)
{
    if (argc < 2)
    {
        std::cout << "Usage: " << argv[0] << " <geo_file>\n";
        return 0;
    }

    gmsh::initialize();

    gmsh::open(argv[1]);
    gmsh::model::mesh::generate(2);

    // Renumber the node tags in a continuous sequence.
    gmsh::model::mesh::renumberNodes(); 

    // output parameters
    std::vector<std::size_t> nodeTags;
    std::vector<double> coord;
    std::vector<double> parametricCoord;

    // gmsh call
    gmsh::model::mesh::getNodes(nodeTags, coord, parametricCoord);


    // display the vectors
    std::cout << "\tnodeTags = ";
    for(int i=0; i<nodeTags.size(); ++i)
        std::cout << nodeTags[i] << " ";
    std::cout << '\n';

    std::cout << "\tcoord = ";
    for(int i=0; i<coord.size(); ++i)
        std::cout << coord[i] << " ";
    std::cout << '\n';   

    gmsh::finalize();
    return 0;
}
